package tvouriot.lib.arbres;

import java.util.Vector ;
import java.util.Enumeration ;
import java.util.Stack ;

/** Classe implantant le type abstrait "arbre d'objets".
 */
public abstract class ArbinObj {
    /** Enumration permettant de rcuprer les donnes d'un arbre.
     */
    private final static class Parcours implements Enumeration {
        /** liste contenant les donnes
         */
        private Vector v ;
        /** Compteur permettant de savoir quel est l'lment courant de l'numration
         */
        private int i ;
        
        /** Constructeur unique
         */
        Parcours() {
            v = new Vector() ;
            i = 0 ;
        }
        
        /** Teste si l'numration  encore des lments
         * @return <CODE>true</CODE> s'il reste des lments, <CODE>false</CODE> sinon
         */
        public boolean hasMoreElements() {
            return i < v.size() ;
        }
        
        /** Renvoie le prochain lment de l'numration
         * @return prochain lment de l'numration
         */
        public Object nextElement() {
            return v.get(i++) ;
        }
    }
    
    /** Retourne la racine de l'arbre courant
     * @return la racine
     */
    public abstract Object racine();
    
    /** Retourne l'arbre gauche de l'arbre courant
     *  @return l'arbre gauche
     */
    public abstract ArbinObj ag();
    
    /** Retourne l'arbre droit de l'arbre courant
     *  @return l'arbre droit
     */
    public abstract ArbinObj ad();
    
    /** Teste si l'arbre est vide
     * @return <CODE>true</CODE> si l'arbre est vide, <CODE>false</CODE> sinon
     */
    public abstract boolean avide();
    
    /** Transforme l'arbre courant en un arbre vide */
    protected abstract void setArbreVide();
    
    /** Retourne un arbre vide
     *  @return arbre vide
     */
    protected abstract ArbinObj getArbreVide();
    
    /** Cr un arbre avec une racine, un arbre gauche et un arbre droit
     *  @param x racine
     *  @param g arbre gauche
     *  @param d arbre droit
     *  @return arbre cr
     */
    protected abstract ArbinObj creationArbre(Object x, ArbinObj g, ArbinObj d);
    
    /** Cr un arbre avec une racine
     *  @param x racine
     *  @return arbre cr
     */
    protected abstract ArbinObj creationArbre(Object x);
    
    /** Renvoie la hauteur de l'arbre courant
     *  @return la hauteur
     */
    public int hauteur() {
        if (this.avide()) return 0;
        return 1+Math.max(this.ad().hauteur(),this.ag().hauteur());
    }
    
    /** Renvoie le nombre de feuilles de l'arbre courant
     *  @return le nombre de feuilles
     */
    public int nbFeuilles() {
        if (this.avide()) return 0;
        if(this.ad().avide() && this.ag().avide()) return 1;
        return this.ad().nbFeuilles() + this.ag().nbFeuilles();
    }
    
    /** Teste si l'arbre est filiforme
     * @return <CODE>true</CODE> si l'arbre est filiforme, <CODE>false</CODE> sinon
     */
    public boolean estFiliforme() {
        if(this.avide()) return true;
        if(this.ag().avide()) return this.ad().estFiliforme();
        if(this.ad().avide()) return this.ag().estFiliforme();
        return false;
    }
    
    /** Retourne le nombre de noeuds de l'arbre courant
     * @return le nombre de noeuds
     */
    public int nbNoeuds() {
        if(this.avide()) return 0;
        return 1+this.ag().nbNoeuds() + this.ad().nbNoeuds();
    }
    
    /** Renvoie une reprsentation sous la forme d'une chane de caractres de l'arbre courant
     *  @return chane de caractres
     */
    public String toString() {
        String st;
        if (this.avide())st=null;
        else {
            if(this.ag().avide() && this.ad().avide())st = "" + this.racine();
            else st = "< " + this.racine() + " , " + this.ag() + " , " + this.ad() + " >";
        }
        return st;
    }
    
    /** Teste si l'arbre courant est complet
     * @return <CODE>true</CODE> si l'arbre est complet, <CODE>false</CODE> sinon
     */
    public boolean estComplet() {
        return this.nbFeuilles() == Math.pow(2, this.hauteur()) ;
    }
    
    /** Teste si l'arbre courant est parfait
     * @return <CODE>true</CODE> si l'arbre est parfait, <CODE>false</CODE> sinon
     */
    public boolean estParfait() {
        int prg, prd ;
        
        if (this.avide())
            return true ;
        else {
            prg = this.ag().hauteur() ;
            prd = this.ad().hauteur() ;
            
            if (prg == prd)
                return this.ag().estComplet() && this.ad().estParfait() ;
            
            if (prg + 1 == prd)
                return this.ag().estParfait() && this.ad().estComplet() ;
            
            return false ;
            
        }
    }
    
    /** Ecrit sur la sortie standart les composants d'un arbre en faisant un parcours prfix
     * @param a arbre  parcourir
     */
    public static void prefixer(ArbinObj a) {
        if (!a.avide()) {
            System.out.print(a.racine() + " - ") ;
            prefixer(a.ag()) ;
            prefixer(a.ad()) ;
        }
    }
    
    /** Ecrit sur la sortie standart les composants d'un arbre en faisant un parcours infix
     * @param a arbre  parcourir
     */
    public static void infixer(ArbinObj a) {
        if (!a.avide()) {
            infixer(a.ag()) ;
            System.out.print(a.racine() + " - ") ;
            infixer(a.ad()) ;
        }
    }
    
    /** Ecrit sur la sortie standart les composants d'un arbre en faisant un parcours postfix
     * @param a arbre  parcourir
     */
    public static void postfixer(ArbinObj a) {
        if (!a.avide()) {
            postfixer(a.ag()) ;
            postfixer(a.ad()) ;
            System.out.print(a.racine() + " - ") ;
        }
    }
    
    /** Utilise une numration pour rcuprer les donnes de l'arbre courant en faisant un parcours prfix
     * @return numration des donnes
     */
    public Enumeration parcoursPre() {
        Parcours p = new Parcours() ;
        parcoursPreBis(p, this) ;
        return p ;
    }
    
    /** Fonction rcursive permettant de faire un parcours prfix
     * @param p numration contenant les lment de l'arbre aprs le parcours
     * @param a arbre  parcourir
     */
    private static void parcoursPreBis(Parcours p, ArbinObj a) {
        if (!a.avide()) {
            p.v.addElement(a.racine()) ;
            parcoursPreBis(p, a.ag()) ;
            parcoursPreBis(p, a.ad()) ;
        }
    }
    
    /** Utilise une numration pour rcuprer les donnes de l'arbre courant en faisant un parcours infix
     * @return numration des donnes
     */
    public Enumeration parcoursInf() {
        Parcours p = new Parcours() ;
        parcoursInfBis(p, this) ;
        return p ;
    }
    
    /** Fonction rcursive permettant de faire un parcours infix
     * @param p numration contenant les lment de l'arbre aprs le parcours
     * @param a arbre  parcourir
     */
    private static void parcoursInfBis(Parcours p, ArbinObj a) {
        if (!a.avide()) {
            parcoursInfBis(p, a.ag()) ;
            p.v.addElement(a.racine()) ;
            parcoursInfBis(p, a.ad()) ;
        }
    }
    
    /** Utilise une numration pour rcuprer les donnes de l'arbre courant en faisant un parcours postfix
     * @return numration des donnes
     */
    public Parcours parcoursPost() {
        Parcours p = new Parcours() ;
        parcoursPostBis(p, this) ;
        return p ;
    }
    
    /** Fonction rcursive permettant de faire un parcours postfix
     * @param p numration contenant les lment de l'arbre aprs le parcours
     * @param a arbre  parcourir
     */
    private static void parcoursPostBis(Parcours p, ArbinObj a) {
        if (!a.avide()) {
            parcoursPostBis(p, a.ag()) ;
            parcoursPostBis(p, a.ad()) ;
            p.v.addElement(a.racine()) ;
        }
    }
    
    /** Utilise une numration pour rcuprer les donnes de l'arbre courant en faisant un parcours prfix (mthode itrative)
     * @return numration des donnes
     */
    public Parcours parcoursPreIter() {
        Parcours p = new Parcours() ;
        Stack pile = new Stack() ;
        ArbinObj a = this ;
        
        while (!a.avide() || !pile.empty()) {
            while(!a.avide()) {
                p.v.addElement(a.racine()) ;
                pile.push(a) ;
                a = a.ag() ;
            }
            
            a = (ArbinObj) pile.pop() ;
            a = a.ad() ;
        }
        
        return p ;
    }
}
